#include <u.h>
#include <libc.h>
#include "httpfs.h"

static char *months[] = {
	"jan", "feb", "mar", "apr", "may", "jun",
	"jul", "aug", "sep", "oct", "nov", "dec"
};

int
exch(Sess *s, char *request, char *hdr, char *fmt, ...)
{
	va_list ap;
	int e, fd;
	char *url, *a[4], buf[128], err[ERRMAX];
	
	va_start(ap, fmt);
	url = vsmprint(fmt, ap);
	va_end(ap);

	if(Dumpreq)
		fprint(2, "request: %s/%q\n", s->url, url);

	/* set headers */
	snprint(buf, sizeof buf, "%s/%d/ctl", s->webfs, s->conid);
	fd = open(buf, OWRITE);
	assert(fd != -1);
	fprint(fd, "url %s%s%s", s->url, (*url == '/')? "": "/", url);
	fprint(fd, "request %s", request);
	if(hdr)
		fprint(fd, "%s", hdr);
	close(fd);

	/* read reply */
	snprint(buf, sizeof buf, "%s/%d/body", s->webfs, s->conid);
	fd = open(buf, OREAD);

	if(fd < 0){
		rerrstr(err, sizeof(err));
		if(tokenize(err, a, nelem(a)) >= 2)
			if((e = atoi(a[1])) > 0)
				httperrstr(e);
		free(url);
		return -1;
	}

	free(url);
	return fd;
}

int
ht_session(Sess *s)
{
	int n, fd;
	char buf[128];

	snprint(buf, sizeof(buf), "%s/clone", s->webfs);
	if((fd = open(buf, ORDWR)) < 0)
		return -1;
	if((n = read(fd, buf, sizeof buf-1)) <= 0){
		close(fd);
		return -1;
	}
	buf[n] = '\0';
	if ((s->conid = atoi(buf)) <0){
		close(fd);
		return -1;
	}
	s->fd = fd;
	return 0;
}

long
ht_get(Sess *s, char *path, char *buf, long len, vlong off)
{
	int fd;
	long got, n;
	char hdr[128];

	snprint(hdr, sizeof(hdr), "headers Range: bytes=%lld-%llud",
		off, (off+len)-1);
	if((fd = exch(s, "GET", hdr, path)) < 0){
		/* NB: we sometimes see the error: "requested range not satisfiable" on EOF */
		return 0; 
	}

	got = 0;
	while(got < len && (n = read(fd, buf+got, len-got)) > 0)
		got += n;
	close(fd);
	return got;
}

long
ht_put(Sess *s, char *path, char *buf, long len, vlong off)
{
USED(s, path, buf, len, off);
/*
	int fd;
	long got, n;
	char hdr[128];

	snprint(hdr, sizeof(hdr), "Content-Range: bytes %lld-%llud/%llud",
		off, (off+len)-1, off+len);
	if((fd = exch(s, "PUT", header, path)) < 0)
		return 0;

	close(fd);
	return got;
*/
	werrstr("not impemented yet");
	return -1;
}


static vlong
contentlength(Sess *s)
{
	int n, fd;
	char buf[1024];

	snprint(buf, sizeof buf, "%s/%d/contentlength", s->webfs, s->conid);
	if((fd = open(buf, OREAD)) < 0)
		return 0;
	n = read(fd, buf, sizeof(buf)-1);
	close(fd);

	if(n < 0)
		return 0;
	buf[n] = 0;
	return strtoll(buf, nil, 0);
}

static long
lastmodified(Sess *s)
{
	Tm tm;
	int i, n, fd;
	char *a[16], buf[8192];

	snprint(buf, sizeof buf, "%s/%d/lastmodified", s->webfs, s->conid);
	if((fd = open(buf, OREAD)) < 0)
		return -1;
	n = read(fd, buf, sizeof(buf)-1);
	close(fd);

	if(n < 0)
		return -1;
	buf[n] = 0;

	if(getfields(buf, a, nelem(a), 1, ", \t:") != 8){
		fprint(2, "%s: not 8 fields, lastmodified header\n", argv0);
		return -1;
	}

	memset(&tm, 0, sizeof(tm));
	for(i = 0; i < nelem(months); i++)
		if(cistrcmp(months[i], a[2]) == 0){
			tm.mon = i;
			break;
		}
	if(i >= nelem(months)){
		fprint(2, "%s: %s bad month, lastmodified header\n", argv0, a[2]);
		return -1;
	}

	tm.mday = atoi(a[1]);
	tm.year = atoi(a[3]) - 1900;
	tm.hour = atoi(a[4]);
	tm.min = atoi(a[5]);
	tm.sec = atoi(a[6]);
	strcpy(tm.zone, a[7]);

	return tm2sec(&tm);
}

long
ht_head(Sess *s, char *path, vlong *length, ulong *mtime)
{
	int fd;

	if((fd = exch(s, "HEAD", nil, path)) < 0)
		return -1;
	*length = contentlength(s);
	*mtime = lastmodified(s);
	close(fd);
	return 0;
}

int
ht_delete(Sess *s, char *path)
{
	int fd;

	if((fd = exch(s, "DELETE", nil, path)) < 0)
		return -1;
	close(fd);
	return 0;
}
